@@ -1,16 +1,17 @@
-import os, inspect, airflow
-from airflow import DAG
+import os, inspect
+from airflow.models import BaseOperator, DAG
 from absql import Runner
 from functools import partial
 from gusty.errors import NonexistentDagDirError
 from gusty.parsing import parse, default_parsers
-from gusty.parsing.loaders import (
-    generate_loader,
-    default_constructors,
-    handle_user_constructors,
-)
+from gusty.parsing.loaders import generate_loader
 from gusty.importing import airflow_version, get_operator
 from gusty.external_sensor import create_external_sensor, make_external_task_name
+from gusty.utils.context import (
+    generate_loader_constructors,
+    generate_runner_context,
+    generate_user_defined_macros,
+)
 
 ###########################
 ## Version Compatability ##
@@ -24,9 +25,9 @@ if airflow_version > 1:
 else:
     from airflow.operators.latest_only_operator import LatestOnlyOperator
 
-#####################
-## Wait for Params ##
-#####################
+###############
+## Constants ##
+###############
 
 AVAILABLE_WAIT_FOR_PARAMS = [
     "poke_interval",
@@ -39,6 +40,8 @@ AVAILABLE_WAIT_FOR_PARAMS = [
     "check_existence",
 ]
 
+BASE_OPERATOR_KEYS = inspect.signature(BaseOperator.__init__).parameters.keys()
+
 #########################
 ## Schematic Functions ##
 #########################
@@ -131,28 +134,34 @@ def parse_wait_for_overrides(external_dependencies):
 #######################
 
 
-def _get_operator_parameters(operator):
-    params = getattr(operator, "_gusty_parameters", None)
-    if params is not None:
+def _get_operator_parameters(operator, operator_param_cache):
+    params = operator_param_cache.get(operator)
+    if params:
         return params
 
-    return inspect.signature(operator.__init__).parameters.keys()
+    params = getattr(operator, "_gusty_parameters", None)
+    if params is not None:
+        operator_param_cache.update({operator: params})
+        return params
+
+    params = inspect.signature(operator.__init__).parameters.keys()
+    operator_param_cache.update({operator: params})
+    return params
 
 
-def build_task(spec, level_id, schematic):
+def build_task(spec, level_id, schematic, operator_cache, operator_param_cache):
     """
     Given a task specification ("spec"), locate the operator and
     instantiate the object with args from the spec.
     """
-    operator = get_operator(spec["operator"])
+    operator = get_operator(spec["operator"], operator_cache)
 
     args = {
         k: v
         for k, v in spec.items()
-        if k
-        in inspect.signature(airflow.models.BaseOperator.__init__).parameters.keys()
-        or k in _get_operator_parameters(operator)
-        or k in _get_operator_parameters(operator.__base__)
+        if k in BASE_OPERATOR_KEYS
+        or k in _get_operator_parameters(operator, operator_param_cache)
+        or k in _get_operator_parameters(operator.__base__, operator_param_cache)
     }
     args["task_id"] = spec["task_id"]
     args["dag"] = get_top_level_dag(schematic)
@@ -241,6 +250,10 @@ class GustyBuilder:
                 "DAG directory {dag_dir} does not exist".format(dag_dir=dag_dir)
             )
 
+        self.extra_tags = kwargs.get("extra_tags")
+
+        self.render_on_create = kwargs.get("render_on_create")
+
         self.parsers = default_parsers.copy()
 
         if len(kwargs["parse_hooks"]) > 0:
@@ -249,20 +262,42 @@ class GustyBuilder:
             ), "parse_hooks should be a dict of file extensions and handler functions for file_path."
             self.parsers.update(kwargs["parse_hooks"])
 
-        # Create dag_constructors
-        dag_constructors = default_constructors.copy()
-        if kwargs["dag_constructors"]:
-            dag_constructors.update(
-                handle_user_constructors(kwargs["dag_constructors"])
-            )
+        # ------------------------- #
+        # Begin Consolidate Context #
+        # ------------------------- #
 
-        # Generate runner with dag_constructors
-        runner_context = {k.strip("!"): v for k, v in dag_constructors.items()}
+        user_defined_macros = kwargs.get("user_defined_macros", {})
+        user_dag_constructors = kwargs.get("dag_constructors", {})
+
+        # Create dag_constructors (for the loader)
+        dag_constructors = generate_loader_constructors(
+            user_defined_macros=user_defined_macros,
+            dag_constructors=user_dag_constructors,
+        )
+
+        # Create runner_context (for the runner)
+        runner_context = generate_runner_context(dag_constructors)
+
+        # Make the runner
         self.runner = Runner(**runner_context)
 
-        # Generate loader with dag_constructors
+        # Make the loader
         self.loader = generate_loader(dag_constructors)
 
+        # Consolidate user_defined_macros
+        kwargs.update(
+            {
+                "user_defined_macros": generate_user_defined_macros(
+                    user_defined_macros=user_defined_macros,
+                    dag_constructors=user_dag_constructors,
+                )
+            }
+        )
+
+        # ----------------------- #
+        # END Consolidate Context #
+        # ----------------------- #
+
         self.schematic = create_schematic(dag_dir, self.parsers)
 
         # DAG defaults - everything that's not task_group_defaults or wait_for_defaults
@@ -286,11 +321,7 @@ class GustyBuilder:
             user_wait_for_defaults = {
                 k: v
                 for k, v in kwargs["wait_for_defaults"].items()
-                if k in AVAILABLE_WAIT_FOR_PARAMS
-                or k
-                in inspect.signature(
-                    airflow.models.BaseOperator.__init__
-                ).parameters.keys()
+                if k in AVAILABLE_WAIT_FOR_PARAMS or k in BASE_OPERATOR_KEYS
             }
             self.wait_for_defaults.update(user_wait_for_defaults)
 
@@ -316,6 +347,10 @@ class GustyBuilder:
         # Allow for the assignment of additional leaf tasks from a dictionary
         self.leaf_tasks_from_dict = kwargs.get("leaf_tasks_from_dict", {})
 
+        # caching
+        self.operator_param_cache = {}
+        self.operator_cache = {}
+
     def parse_metadata(self, id):
         """
         For a given level id, parse any metadata if there is a METADATA.yml path,
@@ -331,7 +366,7 @@ class GustyBuilder:
         level_metadata_path = self.schematic[id]["metadata_path"]
         if os.path.exists(level_metadata_path or ""):
             level_metadata = default_parsers[".yml"](
-                level_metadata_path, self.loader, self.runner
+                level_metadata_path, self.loader, self.runner, render_on_create=True
             )
 
             # special case - default_args provided in both metadata_defaults and level_metadata
@@ -359,6 +394,12 @@ class GustyBuilder:
         else:
             level_metadata = {}
         metadata_defaults.update(level_metadata)
+
+        if self.extra_tags and self.schematic[id]["parent_id"] is None:
+            tags = metadata_defaults.get("tags", [])
+            tags = list(set(tags + self.extra_tags))
+            metadata_defaults.update({"tags": tags})
+
         self.schematic[id]["metadata"] = metadata_defaults
 
         # dependencies get explicity set at the level-"level" for each level
@@ -431,7 +472,9 @@ class GustyBuilder:
         level_metadata = self.schematic[id]["metadata"]
         level_spec_paths = self.schematic[id]["spec_paths"]
         level_specs = [
-            parse(spec_path, self.parsers, self.loader, self.runner)
+            parse(
+                spec_path, self.parsers, self.loader, self.runner, self.render_on_create
+            )
             for spec_path in level_spec_paths
         ]
         level_specs = flatten_nested_lists(level_specs)
@@ -472,7 +515,9 @@ class GustyBuilder:
         """
         level_specs = self.schematic[id]["specs"]
         level_tasks = {
-            spec["task_id"]: build_task(spec, id, self.schematic)
+            spec["task_id"]: build_task(
+                spec, id, self.schematic, self.operator_cache, self.operator_param_cache
+            )
             for spec in level_specs
         }
         self.schematic[id]["tasks"] = level_tasks
